/* thermalbath_control.pde
 *
 * Control firmware for a Haake 000-8560 thermal bath. The core of
 * the control system is an Arduino Pro Mini 5V.
 *
 * The data pins in use are:
 * D0  - TXO (serial, only for computer contact)
 * D1  - RXI (serial, only for computer contact)
 * D2  - pin A encoder
 * D3  - button encoder
 * D4  - pin B encoder
 * D5  - RS pin LCD
 * D6  - E pin LCD
 * D7  - D4 pin LCD
 * D8  - D5 pin LCD
 * D9  - D6 pin LCD
 * D10 - D7 pin LCD
 * D11 - to pump and heater power relay
 * D12 - to solid state relay of heater 
 * D13 - alarm signal in
 * A0  - REGELF temperature sensor from differential amplifier
 * A1  - MESSF temperature sensor from differential amplifier
 * A2  - not connected
 * A3  - not connected
 * A4  - not connected 
 * A5  - not connected
 *
 * It appeared that a P-controller with a gain of K=3000 will do 
 * the trick quite nicely without overshoot or oscillations. 
 * 
 * L. Geers, April 2012
 */

#include <LiquidCrystal.h>
#include <PID_v1.h>
#include <avr/pgmspace.h>
#include <math.h>

// *** LCD strings in program memory *** CHECKED OK
prog_char str_0[] PROGMEM=" Thermo ";
prog_char str_1[] PROGMEM="  v1.0  ";
prog_char str_2[] PROGMEM=" Level  ";
prog_char str_3[] PROGMEM=" Alarm! ";
prog_char str_4[] PROGMEM="Setpoint";
prog_char str_5[] PROGMEM="Press to";
prog_char str_6[] PROGMEM="stop    ";
prog_char str_7[] PROGMEM="Pump    ";
prog_char str_8[] PROGMEM="stopped ";
PROGMEM const char *string_table[]= {
  str_0, str_1, str_2, str_3, str_4, str_5, str_6, str_7, str_8 };

// *** Global constants and variables *** CHECKED OK
LiquidCrystal lcd(5, 6, 7, 8, 9, 10);

const int MINENCVAL= 0;
const int MAXENCVAL= 2000;

volatile int encVal;
int isBtnPressed;
const int encA= 2, 
          encBut= 3,
          encB= 4, 
          pump= 11,
          heat= 12,
          alrm= 13;
char buffer[8], dcnt;
double setpoint, oldregel, inPID, outPID;
unsigned long time, stqtime;

PID tempCtrl(&inPID, &outPID, &setpoint, 3000, 0, 0, DIRECT);
const int windowSize= 5000; // two parameters necessary for PID control
unsigned long windowStartTime; // with on/off actuator

enum Mess { M_SPL= 0, M_VER= 1, M_LVL= 2, M_ALR= 3, M_SP= 4, 
            M_BTP= 5, M_STP= 6, M_PMP= 7, M_STPD= 8 };
enum States {
  S_SPL, // Splash screen
  S_ALR, // Level alarm
  S_CSP, // Change setpoint
  S_RUN, // Running control system
  S_STQ, // Ask to stop
  S_STP  // Pump stopped
};

unsigned char state= S_SPL; // FSM state variable

// *** Request an LCD message from progmem *** CHECKED OK
/* The result is placed in 'buffer', 'm' picks
   the right message. */
void getMsg(unsigned char m) {
  strcpy_P(buffer, (char*)pgm_read_word(&(string_table[m])));
}

// *** Temperature retrieval and calculation function ***
/* Calibration: T= 0.1233*(AD1)^1.1185 (R^2= 0.9995)
 * But this sensor sucks (gives large error) and it is quite slow.
 * Hence, only REGELF will be used.
 */
float getMessF() {
  int v= analogRead(1);
  return 0.1233*pow(v, 1.1185);
}

/* Calibration: T= 0.02941*(AD0)^1.3478 (R^2= 0.9992)
 * Use this one for all measurements and control routines.
 */
float getRegelF() {
  int v= analogRead(0);
  return 0.02941*pow(v, 1.3478);
}

// *** Pin change interrupt routine for alarm *** CHECKED OK
/* Needed a debounce routine, because power surges and sudden dips 
 * cause the ISR to be called. If a trigger appears to be erroneous,
 * the routine should return to the function it was in when called.
 */
ISR(PCINT0_vect) {
  int s=digitalRead(alrm);
  delay(50);
  if ((s & digitalRead(alrm))==HIGH)
    state= S_ALR;
  else state= S_CSP;
}

// *** Encoder turn interrupt service routine *** CHECKED OK
/* If the encoder is turned, the value of encVal will change,
   provided it is still in the working range (MINENCVAL, MAXENCVAL)
*/
void encTurn() {
  if (state!=S_CSP) return; // no change of setting outside setpoint state
  if (digitalRead(encB)==HIGH) {
    if (encVal<MAXENCVAL) encVal++;
  } else {
    if (encVal>MINENCVAL) encVal--;
  }
}

// *** read button routine *** CHECKED OK
boolean getButton() {
  boolean result= false;
  if ((isBtnPressed=!digitalRead(encBut))==HIGH) {
    delay(20);
    if ((isBtnPressed&=!digitalRead(encBut))==HIGH) 
      result= true;
    while (digitalRead(encBut)==LOW);
    delay(50);
  }
  return result;
}

// *** initialization routines *** CHECKED OK
void setup() {
  lcd.begin(8, 2);
  lcd.clear();
  pinMode(encA, INPUT);
  pinMode(encB, INPUT);
  pinMode(encBut, INPUT);
  pinMode(pump, OUTPUT);
  pinMode(heat, OUTPUT);
  pinMode(alrm, INPUT);
  attachInterrupt(0, encTurn, FALLING);
  windowStartTime= millis();
  tempCtrl.SetOutputLimits(0, windowSize); // Set PID PWM controller limits
  tempCtrl.SetMode(MANUAL); // Do not start PID yet.
}

// *** Main loop with finite state machine ***
void loop() {
  float tmp;
  switch (state) {
    case S_SPL:
      lcd.setCursor(0, 0);
      getMsg(M_SPL);
      lcd.print(buffer);
      lcd.setCursor(0, 1);
      getMsg(M_VER);
      lcd.print(buffer);
      delay(2000);
      PCICR |= (1<<PCIE0);
      PCMSK0 |= (1<<PCINT5); // pin change interrupt on alarm
      encVal= (int)(getRegelF()*10);
      oldregel= getRegelF();
      stqtime= 0;
      state= S_CSP;
      break;
    case S_ALR:
      digitalWrite(pump, LOW);
      digitalWrite(heat, LOW);
      tempCtrl.SetMode(MANUAL);
      lcd.setCursor(0, 0);
      getMsg(M_LVL);
      lcd.print(buffer);
      lcd.setCursor(0, 1);
      getMsg(M_ALR);
      lcd.print(buffer);
      break;
    case S_CSP:
      lcd.setCursor(0, 0);
      getMsg(M_SP);
      lcd.print(buffer);
      lcd.setCursor(0, 1);
      tmp= encVal/10.0;
      if (tmp<10.0) lcd.write(' ');
      if (tmp<100.0) lcd.write(' ');
      lcd.print(tmp, DEC);
      lcd.setCursor(5, 1);
      lcd.write(0xDF); // degree symbol
      lcd.print("C ");
      if (getButton()) {
        setpoint= tmp;
        state= S_RUN;
      }
      break;
    case S_RUN:
      if (tempCtrl.GetMode()==MANUAL) {
        tempCtrl.SetMode(AUTOMATIC);
        digitalWrite(pump, HIGH); // alarm interrupt from
        dcnt= 1; // display counter
      }
      tmp= (getRegelF()+oldregel)/2.0; // tone down the fluctuations
      inPID= tmp;
      tempCtrl.Compute();
      time= millis();
      if (time-windowStartTime>windowSize)
        windowStartTime+=windowSize; // shift relay window
      if (outPID>time-windowStartTime)
        digitalWrite(heat, HIGH);
      else
        digitalWrite(heat, LOW);
      oldregel= tmp;
      if (!(--dcnt)) { // display counter to make easier on the eye
        lcd.setCursor(0, 0);
        lcd.write('C');
        if (tmp<10.0) lcd.write(' ');
        if (tmp<100.0) lcd.write(' ');
        lcd.print(tmp, DEC);
        lcd.setCursor(6, 0);
        lcd.write(0xDF); // degree symbol
        lcd.write('C');
        lcd.setCursor(0, 1);
        lcd.write('S');
        if (setpoint<10.0) lcd.write(' ');
        if (setpoint<100.0) lcd.write(' ');
        lcd.print(setpoint, DEC);
        lcd.setCursor(6, 1);
        lcd.write(0xDF); // degree symbol
        lcd.write('C');
        dcnt= 5; // display counter
      }
      if (getButton()) state= S_STQ;
      break;
    case S_STQ:
      lcd.setCursor(0, 0);
      getMsg(M_BTP);
      lcd.print(buffer);
      lcd.setCursor(0, 1);
      getMsg(M_STP);
      lcd.print(buffer);
      if (stqtime==0) stqtime= millis();
      else {
        if (millis()-stqtime>2000) {
          stqtime= 0;
          state= S_RUN;
        }
      }
      if (getButton()) {
        stqtime= 0;
        state= S_STP;
      }
      break;
    case S_STP:
      tempCtrl.SetMode(MANUAL);
      digitalWrite(pump, LOW);
      lcd.setCursor(0, 0);
      getMsg(M_PMP);
      lcd.print(buffer);
      lcd.setCursor(0, 1);
      getMsg(M_STPD);
      lcd.print(buffer);
      delay(2000);
      state= S_CSP;
      break;
  }
  delay(50); // delay to make screen more comfortable to read
}
